#!/usr/bin/env python3
# 
# Cross Platform and Multi Architecture Advanced Binary Emulation Framework
#

from qiling.const import *
from .fncc import *

DWORD = 1
LOCALE_T = 1
UINT = 1
ULONG = 1
INT = 1
BOOL = 1
SIZE_T = 1
BYTE = 1
CCHAR = 1
CHAR = 1
UCHAR = 1
USHORT = 1
SHORT = 1
ULONGLONG = 2
HANDLE = 3
POINTER = 3
STRING = 4
PCANSI_STRING = 4
PANSI_STRING = 4
PCSZ = 4
WSTRING = 5
PCWSTR = 5
PUNICODE_STRING = 6
PCUNICODE_STRING = 6
WCHAR = 7
BOOLEAN = 8
BOOL = 8
STRING_ADDR = 9
WSTRING_ADDR = 10
GUID = 11
